local global = require "eatglobal_v1_17"
local utils = global.utils
local transf = global.transf

local converter = {
	info = {
		version = "0.9",
	},
}

local logFn = function(...) end
local logFnCounter = function(...) end

local transferData = {
	materialNormals = {
		"map_normal",
		"map_dirt_normal",
		"map_rust_normal",
	},
	missingMaterials = {
		"REFLECTIVE",
		"REFLECTIVE_NRML_MAP",
		"REFLECTIVE_OP",
		"REFLECTIVE_NRML_MAP_OP",
		"TRANSPARENT",
		"TRANSPARENT_DETAIL",
		"TRANSPARENT_OP",
		"TRANSPARENT_NRML_MAP",
		"TRANSPARENT_NRML_MAP_OP",
		"PHYSICAL_ALPHA_NRML_MAP",
		"PHYSICAL_ALPHA_NRML_MAP_OP",
	},
	entryNamesData = {
	},
	fileData = {
	},
	loadIndicatorIndizies = {
	},
	userTable = {
		foundedFiles = {},
		maybeDeletedTable = {},
		errorFiles = {},
		usedFiles = {
			meshes = {},
			materials = {},
		},
	},
	other = {
		patternDebugExt = "§ext_debug§ext",
		patternBackupExt = "§ext_backup§ext",
		patternTestExt = "§ext_test§ext",
		fileExt = {".grp", ".mdl", ".mtl", ".msh"},
	},
}

local function dateTimeString()
	return os.date("%d.%m.%Y %X", os.time())
end

local function transferLog(...)
	local s = utils.concatParams(...).."\n"
	utils.saveData(s, transferData.info.transferLogFile, "TEXT", true)
end

local function ixFile(files, fileName)
	for i = 1, #files do
		if string.ends(fileName, files[i]) then
			return i
		end
	end
		
	return 0
end

local function isFile(files, fileName)
	return (ixFile(files, fileName) > 0)
end

local function translMatConfigs(matConfigs)
	if (type(matConfigs) == "table") then
		local result = utils.copyTable(matConfigs)
		--logFnCounter("translMatConfigs | {matConfigs, result} = " , {matConfigs, result})
		for i = 1, #result do
			if (type(result[i]) == "table") then
			for j = 1, #result[i] do
				result[i][j] = result[i][j] + 1
			end
			elseif (type(result[i]) == "number") then
				result[i] = result[i] + 1
			end
		end
		
		return result
	else
		return {}
	end
end

local function getValue(value, default)
	--logFnCounter("getValue | {value, default} = " , {value, default})
	if (type(value) == type(default)) then
		return value
	else
		return default
	end
end

local function transferAnimationToAni(animation, fileName)
	local result = {
		times = {},
		transfs = {},
	}
	
	for i = 1, #animation.params.keyframes do
		local frame = animation.params.keyframes[i]
		result.times[#result.times + 1] = frame.time
		
		local transL = transf.create(animation.params.origin, frame.rot, {0, 0, 0})
		result.transfs[#result.transfs + 1] = transf.setPos(transL, frame.transl)
	end
	
	if (type(fileName) == "string") then
		utils.saveData(result, fileName, "FUNC_DATA")
	end
	
	return result
end

local function getFileExtension(fileName, includeDot)
	local str = fileName
	local temp = ""
	includeDot = (includeDot == nil) and true or includeDot
	local result = includeDot and "." or ""

	for i = str:len(), 1, -1 do
		if str:sub(i, i) ~= "." then
			temp = temp..str:sub(i, i)
		else
			break
		end
	end

	-- Reverse order of full file name
	for j = temp:len(), 1, -1 do
		result = result..temp:sub(j, j)
	end

	return result
end

local function getFileNameAndExtension(fileName)
	local ext = getFileExtension(fileName, false)
	local fName = utils.getFileName(fileName)
	return fName:sub(1, fName:find(ext, 1, true) - 2), ext
end

local function readData(fileName)
	local result, err = utils.readData(fileName, "FUNC_DATA")
	if (err == nil) then
		return result
	else
		--transferLog("Error open file: ", fileName)
		--transferLog(err)
		return nil
	end
end

local function saveData(data, fileName, saveType, mode)
	--logFnCounter("saveData | {fileName, saveType, transferData.debugging} = " , {fileName, saveType, transferData.debugging})
	mode = mode or "FUNC_DATA"
	local extension = getFileExtension(fileName)
	local ext = ""
	if (saveType == "BACKUP") then
		ext = "_backup"..extension
	elseif transferData.debugging then
		ext = "_debug"..extension
	end
	--logFnCounter("saveData | {fileName..ext} = " , {fileName..ext})
	utils.saveData(data, fileName..ext, mode)
end

local function flatStringReaplace(s, pattern, repl)
	local result = s
	local pos = s:find(pattern, 1, true)
	if pos then
		local len = pattern:len()
		if len > 0 then
			result = s:sub(1, pos - 1)..repl..s:sub(pos + len, s:len())
		end
	end
	return result
end

local function clearString(s, prefix, suffix)
	return s:sub(1 + prefix:len(), s:len()- suffix:len())
end

local function getValues(s)
	local result = {}
	
	for k, v in s:gmatch([[(%w+)%s*=%s*([%w+.-]*)]]) do
		result[k] = v
	end
	for k, v in s:gmatch([[(%w+)%s*=%s*([%w+.-]*%b())]]) do
		result[k] = v
	end
	
	return result
end

local function handleTextData(fileName, mode)
	mode = mode or "GET"
	if (type(transferData.fileData[fileName].textData) ~= "table") then
		transferData.fileData[fileName].textData = {}
	end
	local sModel
	if (transferData.debugging and (mode ~= "GET")) then
		local extension = getFileExtension(fileName)
		sModel = utils.readData(fileName.."_debug"..extension, "TEXT")
	else
		sModel = utils.readData(fileName, "TEXT")
	end
	if sModel then
		if (mode == "GET") then
			if transferData.backupFiles then
				saveData(sModel, fileName, "BACKUP", "TEXT")
			end
			-- Description handling
			local sMetadata = string.match(sModel, "metadata%s-=%s-%b{}.-")
			if sMetadata then
				transferData.fileData[fileName].textData.sDesc = string.match(sMetadata, "description%s-=%s-%b{}.-")
				--logFnCounter("handleTextData (GET) | sDesc = ", transferData.fileData[fileName].textData.sDesc)
			end
			
			-- require handling
			local requireString = ""
			for s in string.gmatch(sModel, [[local.-=%s-require%s-%b"".-]]) do
				requireString = requireString..s.."\n"
			end
			transferData.fileData[fileName].textData.requireString = requireString
			
			-- seats handling
			if transferData.oldSeatsStyle then
				transferData.fileData[fileName].textData.seatData = {}
				local seatData = transferData.fileData[fileName].textData.seatData
				local sSeats = string.match(sMetadata, [[seats%s*=%s*%b{}.-]])
				if sSeats then
					sSeats = clearString(string.gsub(sSeats, "seats%s*=%s*", ""), "{", "}")
					for g in string.gmatch(sSeats, [[%s*%b{}.-]]) do
						local t = getValues(g)
						t["standing"] = (t["standing"] == "true") and "true" or "false"
						t["animation"] = (t["standing"] == "true") and [["idle"]] or [["sitting"]]
						t["crew"] = (t["crew"] == "true") and "true" or "false"
						t["forward"] = (t["forward"] == "true") and "true" or "false"
						local sRes = "{ "
						for k, v in utils.sortPairs(t) do
							sRes = sRes..k.." = "..v..", "
						end
						seatData[#seatData + 1] = "\9\9\9\9"..sRes.."},\n"
					end
				end
			end
		else
			local doSave = false
			-- Description handling
			if transferData.fileData[fileName].textData.sDesc then
				local sMetadata = string.match(sModel, "metadata%s-=%s-%b{}.-")
				if sMetadata then
					local sDesc = string.match(sMetadata, "description%s-=%s-%b{}.-")
					if sDesc then
						sModel = flatStringReaplace(sModel, sDesc, transferData.fileData[fileName].textData.sDesc)
						doSave = true
					end
				end
			end
			
			-- require handling
			if transferData.fileData[fileName].textData.requireString then
				sModel = transferData.fileData[fileName].textData.requireString..sModel
				doSave = true
			end
			
			-- seats handling
			if transferData.oldSeatsStyle then
				local sNewSeats = "seats = {\n"..table.concat(transferData.fileData[fileName].textData.seatData).."\9\9\9}"
				if sNewSeats then
					local sSeats = string.match(sModel, [[seats%s*=%s*%b{}.-]])
					if sSeats then
							sModel = flatStringReaplace(sModel, sSeats, sNewSeats)
						doSave = true
					end
				end
			end

			if doSave then
				saveData(sModel, fileName, "", "TEXT")
			end
		end
	end
end

---	transferMaterials
--		params:	-	materials
--						-	
local function transferMaterials(params)
	local function isType(tbl, value)
		for i = 1, #tbl do
			if (value == tbl[i]) then
				return true
			end
		end
		return false
	end
	--logFnCounter("transferMaterials | params = " , params)
	local function handleReflective(data)
		local transparentData = {
			{
				"TRANSPARENT",
				"TRANSPARENT_DETAIL",
				"TRANSPARENT_OP",
			},
			{
				"TRANSPARENT_NRML_MAP",
				"TRANSPARENT_NRML_MAP_OP",
			},
			{
				"PHYSICAL_ALPHA_NRML_MAP",
				"PHYSICAL_ALPHA_NRML_MAP_OP",
			},
		}
		local templates = require "eat1963_tpf2_converter_materials_templates"
		local newData = nil
		if (data.type == "REFLECTIVE") then
			newData = utils.copyTable(templates["PHYSICAL"])
			newData.params.map_albedo.fileName = data.params.map_color_reflect.fileName
			
		elseif (data.type == "REFLECTIVE_NRML_MAP") then
			newData = utils.copyTable(templates["PHYSICAL_NRML_MAP"])
			newData.params.map_albedo.fileName = data.params.map_color_reflect.fileName
			newData.params.map_normal.fileName = data.params.map_normal.fileName
			
		elseif (data.type == "REFLECTIVE_OP") then
			newData = utils.copyTable(templates["PHYSICAL_NRML_MAP_OP"])
			newData.params.map_albedo.fileName = data.params.map_color_reflect.fileName
			
		elseif (data.type == "REFLECTIVE_NRML_MAP_OP") then
			newData = utils.copyTable(templates["PHYSICAL_NRML_MAP_OP"])
			newData.params.map_albedo.fileName = data.params.map_color_reflect.fileName
			newData.params.map_normal.fileName = data.params.map_normal.fileName
			
		elseif isType(transparentData[1], data.type) then
			newData = utils.copyTable(templates["PHYS_TRANSPARENT"])
			newData.params.map_albedo_opacity.fileName = data.params.map_color_alpha.fileName
			
		elseif isType(transparentData[2], data.type) then
			newData = utils.copyTable(templates["PHYS_TRANSPARENT_NRML_MAP"])
			newData.params.map_albedo_opacity.fileName = data.params.map_color_alpha.fileName
			newData.params.map_normal.fileName = data.params.map_normal.fileName
			
		elseif isType(transparentData[3], data.type) then
			newData = utils.copyTable(templates["PHYS_TRANSPARENT_NRML_MAP"])
			newData.params.map_albedo_opacity.fileName = data.params.map_albedo_alpha.fileName
			newData.params.map_normal.fileName = data.params.map_normal.fileName
		end
		
		return newData or data
	end
	
	local function doNormal(data)
		if (type(data) == "table") then
			if (type(data.redGreen) ~= "boolean") then
				data.redGreen = true
				return true
			end
		end
		return false
	end
	
	for i = 1, #params.materials do
		local fileName = transferData.info.materialPath..params.materials[i]
		local data = readData(fileName)
		if (type(data) == "table") then
			local origData = utils.copyTable(data)
			local dataChanged = false
			if isType(transferData.missingMaterials, data.type) then
				data = handleReflective(data)
				--logFn("transferMaterials | {fileName, data} = " , {fileName, data})
				dataChanged = true
			else
				for j = 1, #transferData.materialNormals do
					if doNormal(data.params[transferData.materialNormals[j]]) then
						dataChanged = true
					end
				end
			end
			if dataChanged then
				saveData(data, fileName)
				if transferData.backupFiles then
					saveData(origData, fileName, "BACKUP")
				end
			end
		end
	end
end

---	transferMesh
--		params:	-	fileName
--						-	backupFiles
--						-	matConfigs
local function transferMesh(params)
	--logFnCounter("transferMesh | params = " , params)
	local result = { }
	local function prepareMaterials(subMeshes)
		local result = {}
		for i = 1, #subMeshes do
			if (type(subMeshes[i].materials) == "table") then
				--[[
				-- Material-Dateien transferieren
				transferMaterials( {
						materials = subMeshes[i].materials,
					} )
				]]
				result[#result + 1] = subMeshes[i].materials
				-- Materialien aus Submesh löschen
				subMeshes[i].materials = nil
			else
				return {}, false
			end
		end
		
		return result, true
	end
	
	local function checkData(data)
		local origData = utils.copyTable(data)
		if ((type(data.subMeshes) == "table") and (type(data.matConfigs) == "table"))then
			local materials, res = prepareMaterials(data.subMeshes)
			if res then
				-- Tabellen erstellen und matConfigs/Materialien sichern
				if (type(transferData.fileData[params.fileName]) ~= "table") then
					transferData.fileData[params.fileName] = {}
				end
				transferData.fileData[params.fileName].materials = materials
				transferData.fileData[params.fileName].matConfigs = translMatConfigs(data.matConfigs)
				transferData.fileData[params.fileName].animations = data.animations
				-- matConfigs aus Mesh löschen
				data.matConfigs = nil
				-- Animationen aus Mesh löschen
				data.animations = nil
				
				-- Änderungen speichern
				saveData(data, params.fileName)
				--utils.saveData(data, params.fileName, "FUNC_DATA")
				
				return true, origData
			else
				return false, nil
			end
		else
			if (type(transferData.fileData[params.fileName]) == "table") then
				return true, nil
			else
				return false, nil
			end
		end
	end
	
	local function getMaterials(matConfigIndex)
		local result = {}
		local fileData = transferData.fileData[params.fileName]
		local matConfig = getValue(fileData.matConfigs[matConfigIndex], getValue(fileData.matConfigs[1], {}))
		
		for i = 1, #fileData.materials do
			local material = getValue(fileData.materials[i][matConfig[i]], fileData.materials[i][1])
			result[#result + 1] = material
			transferData.userTable.usedFiles.materials[transferData.info.materialPath..material] = true
		end
		
		return result
	end
	
	-- Hier geht's los :)
	local data = readData(params.fileName)
	if (type(data) == "table") then
		local res, origData = checkData(data)
		if res then
			result.animations = transferData.fileData[params.fileName].animations
			result.materials = getMaterials(params.matConfigIndex)
			
			if (transferData.backupFiles and (type(origData) == "table")) then
				saveData(data, params.fileName, "BACKUP")
				--utils.saveData(data, params.fileName.."_backup.msh", "FUNC_DATA")
			end
		end

		return result
	else
		transferData.userTable.errorFiles[params.fileName] = "No data found. The file may not be available."
		return nil
	end
end

--#########################################################
local function getAnimations(group, events)
	local result = {}
	local found = 0
	
	--logFn("getAnimations | group = " , group)
	if (type(group.animations) == "table") then
		for keyEvent, valueEvent in pairs(events) do
			for keyAnimation, valueAnimation in pairs(group.animations) do
				if (valueEvent.name == keyAnimation) then
					local item = utils.copyTable(valueAnimation)
					item.forward = valueEvent.forward
					result[keyEvent] = item
					found = found + 1
				end
			end
		end
	end
	--logFn("getAnimations | result = " , result)
	return (found > 0) and result or nil
end

local function handleGroupEvents(params)
	local function makeGroupEvents(groupIndex, childEvents, parentEvents)
		local result = {}
		if not utils.isEmptyTable(parentEvents) then
			for keyParent, valueParent in pairs(parentEvents) do
				local event = childEvents[valueParent.name]
				if (type(event) == "table") then
					local item = utils.copyTable(event[groupIndex])
					if (type(item) == "table") then
						item.forward = (valueParent.forward == item.forward) and true or false
						result[keyParent] = item
					end
				end
			end
		else
			for keyChildren, valueChildren in pairs(childEvents) do
				local item = utils.copyTable(valueChildren[groupIndex])
				if (type(item) == "table") then
					result[keyChildren] = item
				end
			end
		end
		
		return result
	end
	
	--*************************************
	
	-- Hier geht's los :)
	local result = {
			animations = {},
			childrenEvents = {},
		}
	local groupEvents = params.group.events or {}
	
	
	if (type(params.group.events) == "table") and (type(params.group.animations) == "table") then
		local gEvents = makeGroupEvents(0, groupEvents, params.parentEvents or {})
		result.animations = getAnimations(params.group, gEvents)
	end
	
	for i, v in ipairs(params.group.children) do
		result.childrenEvents[i] = makeGroupEvents(i, groupEvents, params.parentEvents or {})
	end
	--logFn("handleGroupEvents | {fileName, result} = " , {params.fileName, result})
	return result
end

--#########################################################

local function init(params)
	-- Info, Paths and Files
	transferData.info = utils.copyTable(converter.info)
	transferData.info.modPath = utils.getModPath()
	transferData.info.debugFile = transferData.info.modPath.."converter_debug.lua"
	transferData.info.transferLogFile = transferData.info.modPath.."transfer_log.lua"
	transferData.info.modelsPath = transferData.info.modPath.."res/models/"
	transferData.info.materialPath = transferData.info.modelsPath.."material/"
	transferData.info.meshPath = transferData.info.modelsPath.."mesh/"
	transferData.info.groupPath = transferData.info.modelsPath.."group/"
	transferData.info.modelPath = transferData.info.modelsPath.."model/"
	transferData.info.texturePath = transferData.info.modPath.."res/textures/models/"
	
	transferData.lifespanFactor = 730.5
	transferData.speedFactor = 3.6
		
	-- Flags
	transferData.backupFiles = getValue(params.backupFiles, false)
	transferData.zeroEmissions = getValue(params.zeroEmissions, false)
	transferData.oldSeatsStyle = getValue(params.oldSeatsStyle, false)
	transferData.debugLog = getValue(params.debugLog, false)
	if transferData.debugLog then
		logFn = utils.createLogFn ("[eat1963_tpf2_converter]")
		logFnCounter = utils.createLogFnCounter ("[eat1963_tpf2_converter]")
	end
	transferData.debugGroups = getValue(params.debugGroups, {})
	transferData.debugging = getValue(params.debugging, false)
	
	transferData.handleNoLods = getValue(params.handleNoLods, false)

	-- resetting transfer-log
	utils.saveData("", transferData.info.transferLogFile, "TEXT")
	-- write startup to transfer-log
	transferLog("Startup at "..dateTimeString())
	
	-- search files
	transferData.userTable.foundedFiles.groups = utils.listFiles(transferData.info.groupPath, true, {".grp"})
	transferData.userTable.foundedFiles.models = utils.listFiles(transferData.info.modelPath, true, {".mdl"}, {".mdl_debug.mdl", ".mdl_test.mdl"})
	transferData.userTable.foundedFiles.materials = utils.listFiles(transferData.info.materialPath, true, {".mtl"}, {".mtl_debug.mtl", ".mtl_test.mtl"})
	transferData.userTable.foundedFiles.meshes = utils.listFiles(transferData.info.meshPath, true, {".msh"}, {".msh_debug.msh", ".msh_test.msh"})
	transferData.userTable.foundedFiles.textures = utils.listFiles(transferData.info.texturePath, true)
	
	-- files to be deleted
	--[[
	local whiteList = {}
	for i = 1, #transferData.other.fileExt do
		whiteList[#whiteList + 1] = transferData.other.patternDebugExt:gsub("§ext", transferData.other.fileExt[i])
		whiteList[#whiteList + 1] = transferData.other.patternTestExt:gsub("§ext", transferData.other.fileExt[i])
	end
	transferData.userTable.maybeDeletedTable.groups = utils.listFiles(transferData.info.groupPath, true, whiteList)
	transferData.userTable.maybeDeletedTable.models = utils.listFiles(transferData.info.modelPath, true, whiteList)
	transferData.userTable.maybeDeletedTable.materials = utils.listFiles(transferData.info.materialPath, true, whiteList)
	transferData.userTable.maybeDeletedTable.meshes = utils.listFiles(transferData.info.meshPath, true, whiteList)
	]]
	-- output environment to transfer-log
	transferLog("Info = ", transferData.info)
	transferLog("Founded Files = ", transferData.userTable.foundedFiles)
end

---	converter.transferModels(params)
--		params:	-	models (table)
--						-	backupFiles (bool)
--						-	zeroEmissions (bool)
function converter.transferModels(params)
	local function getEntryName(fileName)
		local ix = getValue(transferData.entryNamesData[fileName], 0) + 1
		transferData.entryNamesData[fileName] = ix
		local fName, fExt = getFileNameAndExtension(fileName)
		return fExt:upper().."_"..fName.."_"..tostring(ix)
	end
	
	local function getLoadIndicatorIndizies(capacities)
		local function doLoadIndicator(loadIndicator)
			if (loadIndicator.type == "LEVEL") then
				local meshIndex = tostring(loadIndicator.params.id)
				local ix = getValue(transferData.loadIndicatorIndizies[meshIndex], 0) + 1
				transferData.loadIndicatorIndizies[meshIndex] = ix
			elseif (loadIndicator.type == "DISCRETE") then
				local levels = getValue(loadIndicator.params.levels, {})
				for i = 1, #levels do
					for j = 1, #levels[i] do
						local meshIndex = tostring(levels[i][j])
						local ix = getValue(transferData.loadIndicatorIndizies[meshIndex], 0) + 1
						transferData.loadIndicatorIndizies[meshIndex] = ix
					end
				end
			end
		end
		
		transferData.loadIndicatorIndizies = {}
		for i = 1, #capacities do
			if ((type(capacities[i].loadIndicator) == "table") and (type(capacities[i].loadIndicator.params) == "table")) then
				doLoadIndicator(capacities[i].loadIndicator)
			end
		end
	end
	
	local function handleMetadata(data)
		local function doEmission(metadata)
			local values = {[true] = 0, [false] = -1}
			if ((type(metadata.transportVehicle) == "table") or (type(metadata.car) == "table")) then
				metadata.emission = {
					idleEmission = values[transferData.zeroEmissions],
					speedEmission = values[transferData.zeroEmissions],
					powerEmission = values[transferData.zeroEmissions],
				}
			end
		end
		if (type(data.metadata) == "table") then
			-- Emissions
			doEmission(data.metadata)
			
			if (type(data.metadata.transportVehicle) == "table") then
				-- Seats
				if (type(data.metadata.transportVehicle.seats) == "table") then
					local tv = data.metadata.transportVehicle
					
					if (type(data.metadata.seatProvider) ~= "table") then
						local sp = { }
						sp.seats = tv.seats
						if tv.carrier == "AIR" then sp.drivingLicense = "AIR" end
						if tv.carrier == "RAIL" then sp.drivingLicense = "RAIL" end
						if tv.carrier == "ROAD" then sp.drivingLicense = "BUS" end
						if tv.carrier == "TRAM" then sp.drivingLicense = "TRAM" end
						if tv.carrier == "WATER" then sp.drivingLicense = "WATER" end
						sp.crewModels = tv.crewModels
						
						for i, seat in ipairs(sp.seats) do
							if seat.standing then
								seat.animation = "idle"
							else
								seat.standing = false
								seat.animation = "sitting"
								-- transf
								if (seat.transf[1] < 0) then
									seat.transf = transf.incrementPos(seat.transf, {0, -0.15, 0})
								else
									seat.transf = transf.incrementPos(seat.transf, {0, 0.15, 0})
								end
							end
							if (type(seat.crew) ~= "boolean") then
								seat.crew = false
							end
							if (type(seat.forward) ~= "boolean") then
								seat.forward = true
							end
						end
						data.metadata.seatProvider = sp
					end
					
					data.metadata.transportVehicle.seats = nil
				end
				
				-- compartments
				if ((type(data.metadata.transportVehicle.capacities) ~= "table") and (type(data.metadata.transportVehicle.compartments) ~= "table")) then
					data.metadata.transportVehicle.compartments = {{{}}}
				elseif (type(data.metadata.transportVehicle.capacities) == "table") then
					local capacities = utils.copyTable(data.metadata.transportVehicle.capacities)
					getLoadIndicatorIndizies(capacities)
					--transferLog("handleMetadata | loadIndicatorIndizies = " , transferData.loadIndicatorIndizies)
					-- delete loadIndicators
					for i = 1, #capacities do
						capacities[i].loadIndicator = nil
					end
					data.metadata.transportVehicle.compartments = {{ capacities }}
					data.metadata.transportVehicle.capacities = nil
				end
				-- Multiple Unit
				if (type(data.metadata.transportVehicle.multipleUnitOnly) ~= "boolean") then
					data.metadata.transportVehicle.multipleUnitOnly = false
				end
				-- Reversible
				if (type(data.metadata.transportVehicle.reversible) ~= "boolean") then
					data.metadata.transportVehicle.reversible = false
				end
			end
			-- other
			if (type(data.metadata.maintenance) == "table") then
				data.metadata.maintenance.lifespan = getValue(data.metadata.maintenance.lifespan, 40) * transferData.lifespanFactor
				data.metadata.maintenance.runningCosts = -1
			end
			data.metadata.cost = getValue(data.metadata.cost, {})
			data.metadata.cost.price = -1
			
			local v
			if (type(data.metadata.railVehicle) == "table") then
				v = data.metadata.railVehicle
				v.topSpeed = v.topSpeed / transferData.speedFactor
				v.configs = getValue(v.configs, {})
				for i = 1, #v.configs do
					v.configs[i].backBackwardParts = getValue(v.configs[i].backBackwardParts, {})
					v.configs[i].backForwardParts = getValue(v.configs[i].backForwardParts, {})
					v.configs[i].blinkingLights0 = getValue(v.configs[i].blinkingLights0, {})
					v.configs[i].blinkingLights1 = getValue(v.configs[i].blinkingLights1, {})
					v.configs[i].frontBackwardParts = getValue(v.configs[i].frontBackwardParts, {})
					v.configs[i].frontForwardParts = getValue(v.configs[i].frontForwardParts, {})
					v.configs[i].innerBackwardParts = getValue(v.configs[i].innerBackwardParts, {})
					v.configs[i].innerForwardParts = getValue(v.configs[i].innerForwardParts, {})
				end
			elseif (type(data.metadata.roadVehicle) == "table") then
				v = data.metadata.roadVehicle
				v.topSpeed = v.topSpeed / transferData.speedFactor
				v.configs = getValue(v.configs, {})
				for i = 1, #v.configs do
					v.configs[i].blinkLightsLeft = getValue(v.configs[i].blinkLightsLeft, {})
					v.configs[i].blinkLightsRight = getValue(v.configs[i].blinkLightsRight, {})
					v.configs[i].brakeLights = getValue(v.configs[i].brakeLights, {})
					v.configs[i].headLights = getValue(v.configs[i].headLights, {})
					v.configs[i].steeringParts = getValue(v.configs[i].steeringParts, {})
				end
			elseif (type(data.metadata.waterVehicle) == "table") then
				v = data.metadata.waterVehicle
				v.waterLine = getValue(v.waterLine {})
				v.type = (v.weight <= 250000) and "SMALL" or "BIG"
				v.configs = getValue(v.configs, {})
			elseif (type(data.metadata.airVehicle) == "table") then
				v = data.metadata.airVehicle
				v.configs = getValue(v.configs, {})
				for i = 1, #v.configs do
					v.configs[i].beaconLights = getValue(v.configs[i].beaconLights, {})
					v.configs[i].landingLight = getValue(v.configs[i].landingLight, {})
					v.configs[i].steeringParts = getValue(v.configs[i].steeringParts, {})
					v.configs[i].strobeLights = getValue(v.configs[i].strobeLights, {})
					if (type(v.configs[i].props) == "table") then
						local ids = utils.copyTable(v.configs[i].props.ids)
						v.configs[i].props = getValue(ids, {})
					end
				end
				v.type = (v.wingArea <= 140) and "SMALL" or "BIG"
			end
			
			if (type(v) == "table") then
				v.engines = getValue(v.engines, {})
			end
		end
	end
	
	local function handleMesh(params)
		if (transferData.loadIndicatorIndizies[tostring(params.lodData.currentOriginalID)] == nil) then
			local children = params.group.children[params.index]
			local fileName = transferData.info.meshPath..children.id
			local data = transferMesh( {fileName = fileName, matConfigIndex = params.matConfigIndex})
			if (type(data) == "table") then
				params.lodData.currentMeshID = params.lodData.currentMeshID + 1
				local result = {
					_meshId = params.lodData.currentMeshID,
					_origMeshId = params.lodData.currentOriginalID,
					name = getEntryName(fileName),
					mesh = children.id,
					transf = children.transf,
					materials = data.materials,
					animations = getAnimations(data, params.parentEvents),
				}
				table.insert(params.root, result)
				transferData.userTable.usedFiles.meshes[fileName] = true
			else
				transferData.userTable.errorFiles[fileName] = "No data found. The file may not be available."
			end
		end
	
	end

	local function handleGroup(params)
		local groupEvents = handleGroupEvents(params)
		--logFn("handleGroup | groupEvents = ", groupEvents)
		params.groupRoot.animations = groupEvents.animations
		
		
		params.lodData.currentID = params.lodData.currentID + 1
		for i = 1, #params.group.children do
			--logFn("handleGroup | {number, id} = ", {i, params.group.children[i].id})
			local groupConfig = translMatConfigs(params.group.matConfigs)
			--local mConfigs = translMatConfigs(getValue(params.group.matConfigs[getValue(params.matConfigs[i], 1)], {}))
			local mConfigs = getValue(groupConfig[getValue(params.matConfigs[params.index], 1)], {})
			params.lodData.currentOriginalID = params.lodData.currentOriginalID + 1
			
			if (params.group.children[i].type == "MESH") then
				handleMesh( {
						--fileName = params.group.children[i].id,
						root = params.root,
						group = params.group,
						index = i,
						parentEvents = groupEvents.childrenEvents[i],
						matConfigIndex = getValue(mConfigs[i], 1),
						lodData = params.lodData,
					} )
			elseif (params.group.children[i].type == "GROUP") then
				local data = readData(transferData.info.groupPath..params.group.children[i].id)
				if (type(data) == "table") then
					params.lodData.currentMeshID = params.lodData.currentMeshID + 1
					local groupEntry = {
						children = {},
						name = getEntryName(params.group.children[i].id),
						transf = params.group.children[i].transf,
					}
					handleGroup( {
							fileName = params.group.children[i].id,
							root = groupEntry.children,
							groupRoot = groupEntry,
							group = data,
							index = i,
							parentEvents = groupEvents.childrenEvents[i],
							matConfigs = mConfigs,
							lodData = params.lodData,
						} )
						if (#groupEntry.children > 0) then
							params.root[#params.root + 1] = groupEntry
						else
							params.lodData.currentMeshID = params.lodData.currentMeshID - 1
						end
				else
					transferData.userTable.errorFiles[transferData.info.groupPath..params.group.children[i].id] = "No data found. The file may not be available."
				end
				if transferData.debugLog and isFile(transferData.debugGroups, params.group.children[i].id) then
					utils.saveData(params.groupRoot, transferData.info.debugFile, "RETURN")
				end
			end
		end
	end
	
	-- Hier geht's los :)
	params = params or {}
	init(params)
	--transferLog("Start Convert: ", dateTimeString())
	local models = params.models or transferData.userTable.foundedFiles.models
	for i = 1, #models do 
		local fileName = transferData.info.modelPath..models[i]
		local data = readData(fileName)
		
		if (type(data) == "table") then
			if (getValue(data.version, 0) == 0) then
				transferData.fileData[fileName] = {
						lodData = {},
					}
				handleTextData(fileName, "GET")
				-- Lods
				--if (type(data.lods) == "table") and (not transferData.handleNoLods) then
				if (type(data.lods) == "table") then
					handleMetadata(data)
					local newLods = {
					}
					for j = 1, #data.lods do
						newLods[#newLods + 1] = {
							node = {
								name = "RootNode",--utils.getFileName(fileName),
								children = {},
								transf = { 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, },
							},
							static = data.lods[j].static,
							visibleFrom = data.lods[j].visibleFrom,
							visibleTo = data.lods[j].visibleTo,
						}
						transferData.fileData[fileName].lodData[j] = {
								currentID = 0,
								currentMeshID = 0,
								currentOriginalID = 0,
								idTables = { },
							}
						handleGroup( {
								root = newLods[#newLods].node.children,
								groupRoot = newLods[#newLods].node,
								group = data.lods[j],
								index = 1,
								parentEvents = {},
								matConfigs = {},
								lodData = transferData.fileData[fileName].lodData[j],
							} )
					end
					data.lods = newLods
					
					-- collider
					if (type(data.collider) ~= "table") then
						data.collider = {
							params = { },
							transf = { 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, },
							type = "MESH",
						}
					end
					if (type(data.collider.params) ~= "table") then
						data.collider.params = {}
					end
					if (type(data.collider.transf) ~= "table") then
						data.collider.transf = { 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, }
					end
					if (data.collider.type ~= "MESH") then
						data.collider.type = "MESH"
					end
					
					-- version
					data.version = 1
					
					-- speichern
					saveData(data, fileName)
					
					--
					handleTextData(fileName, "SET")
				end
			end
		end
		--logFn("transferModels | {fileName, data, newLods} = ", {fileName, data, newLods})
	end
	transferMaterials( {
		materials = transferData.userTable.foundedFiles.materials,
	} )
	
	--[[
	if (not transferData.debugging) then
		local files = {}
		for i = 1, #transferData.userTable.maybeDeletedTable.groups do
			files[#files + 1] = transferData.userTable.maybeDeletedTable.groups[i]
		end
		for i = 1, #transferData.userTable.maybeDeletedTable.models do
			files[#files + 1] = transferData.userTable.maybeDeletedTable.models[i]
		end
		for i = 1, #transferData.userTable.maybeDeletedTable.materials do
			files[#files + 1] = transferData.userTable.maybeDeletedTable.materials[i]
		end
		for i = 1, #transferData.userTable.maybeDeletedTable.meshes do
			files[#files + 1] = transferData.userTable.maybeDeletedTable.meshes[i]
		end
		transferLog("Delete Files ", files)
		for i = 1, #files do
			utils.deleteFile(transferData.info.modPath..files[i])
		end
	end
	]]
	transferLog("Used Files = ", transferData.userTable.usedFiles)
	transferLog("Error Files = ", transferData.userTable.errorFiles)
	transferLog("Finished at "..dateTimeString())
end

function converter.transferMaterials(params)
	init(params)
	transferMaterials(params)
	transferLog("Finished at "..dateTimeString())
end

--[[
	Erstellt auf Grundlage eines TPF2-Models (masterFile) Repaint-Models.
	Hierbei werden die entsprechenden Repaintfiles erzeugt, die angegebenen Material-Einträge
	geändert und das masterFile zur Unterstützung der Depot-Repaint-Function eingetragen.
	
	
	converter.createRepaints(params)
		params:	-	data "table" (siehe Beschreibung weiter unten)
						-	debugLog "boolean"
						-	debugging "boolean"
	-----------------------------------------------------
	Aufbau table params.data (mit Beispieleinträgen):
	repaintData = {
		masterFile = "vehicle/train/masterModel.mdl", -- dient als Basis
		createMaterialFiles = false, -- sollen Repaint-Materialdateien erstellt werden?
		repaintFiles = {
			-- Datei wird erstellt
			["vehicle/train/repaintModel_01.mdl"] = {
				{vehicle/waggon/masterMaterial_01.mtl, vehicle/waggon/repainMaterial_01_01.mtl}, -- links Original Material; rechts Repaint-Material
				{vehicle/waggon/masterMaterial_02.mtl, vehicle/waggon/repainMaterial_01_02.mtl}, -- links Original Material; rechts Repaint-Material
				{vehicle/waggon/masterMaterial_03.mtl, vehicle/waggon/repainMaterial_01_03.mtl}, -- links Original Material; rechts Repaint-Material
			},
			-- Datei wird ebenfalls erstellt
			["vehicle/train/repaintModel_02.mdl"] = {
				{vehicle/waggon/masterMaterial_01.mtl, vehicle/waggon/repainMaterial_02_01.mtl}, -- links Original Material; rechts Repaint-Material
				{vehicle/waggon/masterMaterial_02.mtl, vehicle/waggon/repainMaterial_02_02.mtl}, -- links Original Material; rechts Repaint-Material
				{vehicle/waggon/masterMaterial_03.mtl, vehicle/waggon/repainMaterial_02_03.mtl}, -- links Original Material; rechts Repaint-Material
			},
		},
	}
	-----------------------------------------------------
	Aufruf:
		converter.createRepaints( {
			data = repaintData,
			createMaterialFiles = true/false,
		} )
]]
function converter.createRepaints(params)
	local function doDepotRepaint(fileName)
		local data, err = utils.readData (fileName, "FUNC_DATA")
		if (err == nil) then
			data.metadata.groupFileName = params.data.masterFile
			utils.saveData(data, fileName, "FUNC_DATA")
		end
	end
	
	init(params)
	if (type(params.data) == "table") then
		local fileName = transferData.info.modelPath..params.data.masterFile
		local masterText, err = utils.readData(fileName, "TEXT")
		if (masterText ~= nil) then
			for repaintFile, repaintData in pairs(params.data.repaintFiles) do
				transferLog("createRepaints | work on '", transferData.info.modelPath..repaintFile, "'")
				local repaintText = masterText
				for i = 1, #repaintData do
					repaintText = repaintText:gsub(repaintData[i][1], repaintData[i][2])
					
					if (params.createMaterialFiles) then
						local mFileDest = transferData.info.materialPath..repaintData[i][2]
						if (not utils.fileExists(mFileDest)) then
							transferLog("createRepaints | work on '", mFileDest, "'")
							local mFileSource = transferData.info.materialPath..repaintData[i][1]
							local mData = utils.readData(mFileSource, "FUNC_DATA")
							utils.saveData(mData, mFileDest, "FUNC_DATA")
						end
					end
				end
				
				utils.saveData (repaintText, transferData.info.modelPath..repaintFile, "TEXT")
				doDepotRepaint(transferData.info.modelPath..repaintFile)
			end
		else
			transferLog("createRepaints | ", err)
		end
	else
		transferLog("createRepaints | params.data is not a table ", type(params.data))
	end
	transferLog("Finished at "..dateTimeString())
end

return converter